# -*- coding: utf-8 -*-
"""Confusion matrix module.

API:
* *ConfusionMatrix* (class): Forms up confusion matrix.
* *BinaryConfusionMatrix* (class): Binary version of confusion matrix class.

Example:
    Binary Confusion matrix with [False, True] labels:
        >>> from reason.metrics import BinaryConfusionMatrix
        >>> cm = BinaryConfusionMatrix(y_true, y_pred)

        >>> cm
        68 21
        24 87

        >>> cm[True, False]
        24

        >>> cm.accuracy()
        0.775
        >>> cm.precision()
        0.7837
        >>> cm.recall()
        0.8055
        >>> cm.f1_score()
        0.7944

"""


class ConfusionMatrix:
    """Confusion Matrix

    Forms up an confusion matrix with given data. Has precision, recall and
    f1 score for binary cases.

    """

    def __init__(self, y_true, y_pred):
        """ConfusionMatrix constructor.

        Takes input data and makes a dict type matrix for storing and returning
        data and a list type matrix for presentation.

        Args:
            y_true (list): Real labels.
            y_pred (list): Predicted labels returned by classifier.

        Raises:
            Exception: If input data is not valid.

        """
        try:
            self._y_true = list(y_true)
            self._y_pred = list(y_pred)
        except TypeError:
            raise Exception('Inputs must be array-like objects.')

        self._labels = sorted(set(self._y_true))
        self._n_y = len(self._y_true)

        self._matrix = dict()
        for i in self._labels:
            self._matrix[i] = dict()
            for j in self._labels:
                self._matrix[i][j] = 0

        for i in range(self._n_y):
            self._matrix[self._y_pred[i]][self._y_true[i]] += 1

        self._array_matrix = [
            [str(n) for n in self._matrix[key].values()]
            for key in self._matrix.keys()
        ]

    def __str__(self):
        """String converter.

        Columns represent actual values, Rows represent predicted values.

        Returns:
            str: String presentation of the matrix.

        """
        return '\n'.join(
            ' '.join(
                self._array_matrix[i]
            ) for i in range(len(self._labels))
        )

    def __getitem__(self, keys):
        true, pred = keys
        return self._matrix[pred][true]

    def accuracy(self):
        """Accuracy method.

        Returns:
            float: Accuracy.

        """
        correct = 0.0
        for label in self._labels:
            correct += self._matrix[label][label]
        return float('{:.4f}'.format(correct / self._n_y))



class BinaryConfusionMatrix(ConfusionMatrix):
    """Binary Confusion Matrix

    Base class: ConfusionMatrix

    """

    def __init__(self, y_true, y_pred):
        """BinaryConfusionMatrix constructor.

        Calls base class constructor, then checks if matrix is binary (if labels
        set has two values.

        Args:
            y_true (list): Real labels.
            y_pred (list): Predicted labels returned by classifier.

        Raises:
            Exception: If labels set has less or more than two values.

        """
        super().__init__(y_true, y_pred)
        if len(self._labels) != 2:
            raise Exception('Binary matrix should have only two labels.')

    def precision(self):
        """Precision method.

        Returns:
            float: Precision.

        """
        false, true = self._labels[0], self._labels[1]
        return float('{:.4f}'.format(
            self._matrix[true][true] /
            (self._matrix[true][true] + self._matrix[true][false])
        ))

    def recall(self):
        """Recall method.

        Returns:
            float: Recall.

        """
        false, true = self._labels[0], self._labels[1]
        return float('{:.4f}'.format(
                self._matrix[true][true] /
                (self._matrix[true][true] + self._matrix[false][true])
        ))

    def f1_score(self):
        """F1 score method.

        Returns:
            float: F1 score.

        """
        precision, recall = self.precision(), self.recall()
        return float('{:.4f}'.format(
            2 * precision * recall / (precision + recall)
        ))



# -*- coding: utf-8 -*-
"""Frequency distribution module.

API:
* *ngrams* (function): Get input ngrams.
* *bigrams* (function): Get input bigrams.
* *trigrams* (function): Get input trigrams.

Example:
    Get ngrams:
        >>> sent = 'Reason is easy to use'

        >>> from reason.util import bigrams
        >>> bigrams(sent)
        [('Reason', 'is'), ('is', 'easy'), ('easy', 'to'), ('to', 'use')]

        >>> from reason.util import trigrams
        >>> trigrams(sent)
        [('Reason', 'is', 'easy'), ('is', 'easy', 'to'), ('easy', 'to', 'use')]

        >>> from reason.util import ngrams
        >>> ngrams(sent, 4)
        [('Reason', 'is', 'easy', 'to'), ('is', 'easy', 'to', 'use')]

"""
from reason.tokenize import word_tokenize


def ngrams(input, n=1):
    """Ngrams function.

    Args:
        input (str or list of str): Text or corpus.
        n (int): Number.

    Returns:
        list of tuple: Ngrams.

    """
    try:
        data = word_tokenize(input)
    except TypeError:
        raise Exception('Input type is not supported.')

    ngrams = list()
    for i in range(len(data) - (n - 1)):
        ngram = tuple([token for token in data[i : (i + n)]])
        ngrams.append(ngram)

    return ngrams


def bigrams(input):
    """Bigrams function.

    Uses Ngrams with n = 2.

    Args:
        input (str or list of str): Text or corpus.

    Returns:
        list of tuple: Bigrams.

    """
    return ngrams(input, 2)


def trigrams(input):
    """Trigrams function.

    Uses Ngrams with n = 3.

    Args:
        input (str or list of str): Text or corpus.

    Returns:
        list of tuple: Trigrams.

    """
    return ngrams(input, 3)



# -*- coding: utf-8 -*-
"""Word tokenize module.

API:
* *accuracy* (function): Calculates accuracy score.

Example:
    Tokenize an string:

        >>> from reason.metrics import accuracy

        >>> accuracy(y_true, y_pred)
        0.9358

"""


def accuracy(y_true, y_pred):
    """Accuracy score function.

    Easy-to-use word tokenize function.

    Args:
        y_true (list): Real labels.
        y_pred (list): Predicted labels returned by classifier.

    Returns:
        float: Accuracy score.

    """
    all = len(y_true)
    correct = 0
    for i in range(all):
        if y_true[i] == y_pred[i]:
            correct += 1
    return float('{:.4f}'.format(correct / all))



# -*- coding: utf-8 -*-
"""Word tokenize module.

API:
* *NaiveBayesClassifier* (class): Naive bayes classifier.

Example:
    Train a classifier and classify new entries:
        >>> from reason.classify import NaiveBayesClassifier

        >>> classifier = NaiveBayesClassifier(train_set)
        >>> y_pred = classifier.classify(data)

        >>> classifier.get_labels()
        [True, False]
        >>> classifier.get_features()
        ['length', 'is_upper', 'ends_with']

"""
import numpy as np
import pandas as pd

# Supported feature types
_numeric = [int, float, bool, np.intc , np.single, np.bool_]
_categorical = [str]


class NaiveBayesClassifier:
    """Naive Bayes Classifier

    Uses gaussian distribution when dealing with continuous data.

    """

    def train(self, x, y):
        """Train method.

        Trains classifier with dataset.

        Args:
            x (pandas.DataFrame or list of dict): Feature sets
            y (pandas.Series or list): Labels

        Raises:
            Exception: If input data is not valid.

        """
        try:
            self._y = pd.Series(y)
        except TypeError:
            raise Exception('Y must be an array-like object.')
        if type(x) == pd.DataFrame:
            self._x = x
        elif self._is_featuresets_format(x):
            self._x = self._featuresets_to_dataframe(x)
        else:
            raise Exception('Dataset type is not supported.')
        self._dataset = self._x.copy()
        self._dataset['label'] = self._y

        self._train_classifier()

    def classify(self, data):
        """Classify method.

        Classifies new entries (feature sets).

        Args:
            input (pandas.DataFrame or list of dict): Features set(s).

        Returns:
            Label or list of labels.

        Raises:
            Exception: If input is not valid.

        """
        if type(data) == pd.Series:
            return self._classify_data(data)
        elif type(data) == dict:
            return self._classify_data(pd.Series(data))
        elif type(data) == pd.DataFrame:
            x = data
        elif self._is_featuresets_format(data):
            x = self._featuresets_to_dataframe(data)
        else:
            raise Exception('Data type is not supported.')

        labels = list()
        for i in range(len(x)):
            labels.append(self._classify_data(x.iloc[i]))
        return labels

    def get_labels(self):
        """Get labels method.

        Returns:
            List: Labels

        """
        return list(self._labels)

    def get_features(self):
        """Get features method.

        Returns:
            List: Features

        """
        return self._features

    def _train_classifier(self):

        self._n = min(len(self._x), len(self._y))

        self._features = list(self._x.columns)
        self._labels = list(self._y.unique())

        self._prior = dict()
        for label in self._labels:
            self._prior[str(label)] = \
                self._y.value_counts()[label]

        self._statistics = dict()
        for label in self._labels:
            features = dict()
            for feature in self._features:
                if self._x[feature].dtype in _numeric:
                    features[feature] = {
                        'mean': np.mean(
                            self._dataset[
                                self._dataset['label'] == label
                            ][feature]
                        ),
                        'var': np.var(
                            self._dataset[
                                self._dataset['label'] == label
                            ][feature]
                        ),
                    }
            self._statistics[str(label)] = features

    def _classify_data(self, x):
        assert isinstance(x, pd.Series), 'X type must be pandas.Series'
        posterior = list()
        for label in self._labels:
            posterior.append((
                self._prior[str(label)] * self._likelihood(x, label), label
            ))
        return max(posterior)[1]

    def _likelihood(self, x, label):
        p = list()
        for feature in self._features:
            if type(x[feature]) in _numeric:
                mean = self._statistics[str(label)][feature]['mean']
                var = self._statistics[str(label)][feature]['var']
                if var == 0:
                    p.append(1)
                else:
                    p.append(
                        1 / np.sqrt(2 * np.pi * var)
                        * np.exp((-(x[feature] - mean) ** 2) / (2 * var))
                    )

            elif type(x[feature]) in _categorical:
                p.append((
                    self._dataset[
                        self._dataset.iloc[:, -1] == label
                    ][feature].value_counts(x[feature]) / self._n
                )[0])

            else:
                raise Exception('Input type is not supported.')

        return np.prod(p)

    def _is_featuresets_format(self, input):

        if not isinstance(input, list):
            return False
        elif not all(isinstance(item, dict) for item in input):
            return False

        return True

    def _featuresets_to_dataframe(self, featuresets):
        data = dict()
        features = featuresets[0].keys()

        for feature in features:
            data[feature] = pd.Series(set[feature] for set in featuresets)

        df = pd.DataFrame(data=data)

        return df



# -*- coding: utf-8 -*-
"""Word tokenize module.

API:
* *WordTokenizer* (class): For frequent use with an specific regex pattern.
* *word_tokenize* (function): For instant use.

Example:
    Tokenize an string:

        >>> from reason.tokenize import word_tokenize

        >>> sentence = "Testing reason0.1.0, (on: 127.0.0.1). Cool stuff..."
        >>> word_tokenize(sentence, 'alphanumeric')
        ['Testing', 'reason0.1.0', 'on', '127.0.0.1', 'Cool', 'stuff']

"""
import re


re_collection = {
    'etc': r"[\.]{3}",
    'ideogram': r"[:;][\)\(\*\|o0]+|[\)\(\*\|o0][:;]+|[-@$\^\*]_[-@$\^\*]",
    'abbr': r"(?:[A-Z]\.)+",
    'decimal': r"\$?\d+(?:,\d+)*(?:\.\d+)*%?",
    'word': r"\w+(?:['\.]\w+)*",
    'other': r"[^\w\s]",
}

re_micro_collection = {
    'alpha': r"[A-Za-z]+(?:'[A-Za-z]+)?",
    'numeric': r"\d+(?:,\d+)*(?:\.\d+)*",
}

re_pure_collection = {
    'alpha': r"[A-Za-z]+",
    'numeric': r"\d+",
}

regex_patterns = {
    'default': '|'.join([re for re in re_collection.values()]),
    'word': re_collection['word'],
    'ideogram': re_collection['ideogram'],
    'alphanumeric': re_collection['word'],
    'alpha': re_micro_collection['alpha'],
    'alpha-pure': re_pure_collection['alpha'],
    'numeric': re_micro_collection['numeric'],
    'numeric-pure': re_pure_collection['numeric'],
    'non-alphanumeric': '|'.join([
        re_collection['etc'], re_collection['ideogram'], re_collection['other']
    ]),
}


class WordTokenizer:
    """Regex Word Tokenizer

    Attributes:
        pattern (str): Regex pattern for tokenizing.

    """

    def __init__(self, pattern=None):
        """WordTokenizer Constructor.

        Sets a regex pattern for tokenizing.

        Ready-to-use choices for pattern are: 'default', 'word', 'ideogram',
        'alphanumeric', 'alpha', 'alpha-pure', 'numeric', 'numeric-pure',
        'non-alphanumeric'.

        You can also define your own regex pattern in the format of string.

        Args:
            pattern (str, optional): Regex pattern to use for tokenizing

        Raises:
            Exception: If pattern is not a valid regex.

        """
        if pattern == None:
            self.pattern = regex_patterns['default']
        elif pattern in regex_patterns.keys():
            self.pattern = regex_patterns[pattern]
        else:
            try:
                re.compile(pattern)
                self.pattern = pattern
            except:
                raise Exception('Pattern is not valid.')

    def tokenize(self, input):
        """Tokenize text method.

        Tokenize input text

        Args:
            input (str or list of str): Text to tokenize.

        Returns:
            list: Tokens.

        Raises:
            Exception: If input is not string or a list of strings.

        """
        if type(input) == str:
            text = input
        else:
            try:
                text = ' '.join(input)
            except TypeError:
                raise Exception(
                    'Tokenize input must be string or a list of strings.'
                )

        return re.findall(self.pattern, text)


def word_tokenize(input, pattern='default'):
    """Tokenize text function.

    Easy-to-use word tokenize function.

    Args:
        input (str or list of str): Text to tokenize.
        pattern (str, optional): Regex pattern to use for tokenizing.

    Returns:
        list: Tokens.

    """
    return WordTokenizer(pattern).tokenize(input)



# -*- coding: utf-8 -*-
"""Sentence tokenize module.

API:
* *SentTokenizer* (class): For frequent sentence tokenizing use.
* *sent_tokenize* (function): For instant use.

Example:
    Tokenize an string:

        >>> from reason.tokenize import sent_tokenize

        >>> text = "Hey, what's up? I love using Reason library!"
        >>> sents = sent_tokenize(text)
        >>> for sent in sents:
        ...     print(sent)
        Hey, what's up?
        I love using Reason library!

"""
from reason.classify import NaiveBayesClassifier
from ._treebank import get_sents
from .word_tokenize import WordTokenizer


class SentTokenizer:
    """Sentence Tokenizer
    """

    def __init__(self):
        """SentTokenizer Constructor.

        Creates and trains a naive bayes classifier. SentTokenizer class
        recognizes sentences using this classifier.

        """
        self._classifier = NaiveBayesClassifier()
        x, y = self._get_dataset()
        self._classifier.train(x, y)

    def tokenize(self, input):
        """Tokenize text method.

        Tokenize input text

        Args:
            input (str or list of str): Text to tokenize.

        Returns:
            list: Tokens.

        Raises:
            Exception: If input is not string or a list of strings.

        """
        wt = WordTokenizer()
        if type(input) == str:
            words = wt.tokenize(input)
            sents = [input]
            input_type = 'str'
        else:
            try:
                words = wt.tokenize(' '.join(input))
                sents = list()
                input_type = 'list'
            except TypeError:
                raise Exception(
                    'Tokenize input must be string or a list of strings.'
                )

        position, start = 0, 0

        for i, token in enumerate(words):
            try:
                if token in '.?!' and self._classifier.classify(
                    self._punc_features(words, i)
                ) == True:
                    if input_type == 'str':
                        text = sents.pop(-1)
                        sent, rest = text[: position + 1], text[position + 1 :]
                        sents.append(sent.strip())
                        sents.append(rest.strip())
                        position = 0
                    else:
                        sents.append(words[start : i + 1])
                        start = i + 1

                if input_type == 'str':
                    position += len(token)
                    if sents[-1][position] == ' ':
                        position += 1

            except IndexError:
                break

        if input_type == 'list':
            if start < len(words):
                if input_type == 'str':
                    sents.append(words[start:])
                else:
                    sents.append(words[start:])

        return sents

    def _punc_features(self, tokens, i):
        return {
            'next_word_capitalized': tokens[i + 1][0].isupper(),
            'punctuation': tokens[i],
            'prev_word': tokens[i - 1].lower(),
            'prev_word_is_one_char': len(tokens[i - 1]) == 1,
        }

    def _get_dataset(self):
        sents = get_sents()
        tokens = list()
        boundaries = set()
        offset = 0

        for sent in sents:
            tokens.extend(sent)
            offset += len(sent)
            boundaries.add(offset - 1)

        featuresets, labels = list(), list()

        for i in range(1, len(tokens) - 1):
            if tokens[i] in '.?!':
                featuresets.append(
                    self._punc_features(tokens, i)
                )
                labels.append(
                    (i in boundaries)
                )


        return featuresets, labels


def sent_tokenize(input):
    """Tokenize text function.

        Easy-to-use sentence tokenize function.

        Args:
            input (str or list of str): Text to tokenize.

        Returns:
            list: Tokens.

        """
    return SentTokenizer().tokenize(input)



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='reason',
    version='0.3.0',
    description='NLP Toolbox',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/alisoltanirad/reason',
    author='Ali Soltani Rad',
    author_email='soltaniradali@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Natural Language :: English',
    ],
    python_requires='>=3.7',
    install_requires=[
        'numpy'
    ],
)



# -*- coding: utf-8 -*-
"""Frequency distribution module.

API:
* *FreqDist* (class): Frequency distribution.

Example:
    Find word frequencies:
        >>> from reason.analysis import FreqDist

        >>> words = ['hey', 'hey', 'oh', 'oh', 'oh', 'yeah']
        >>> fd = FreqDist(words)

        >>> fd
        Frequency Distribution
        Most-Common: [('oh', 3), ('hey', 2), ('yeah', 1)]
        >>> fd.most_common(2)
        [('oh', 3), ('hey', 2)]
        >>> fd['yeah']
        1

"""
from collections import Counter
from copy import deepcopy

from reason.tokenize import word_tokenize


class FreqDist:
    """Frequency Distribution

    Counts token frequencies.

    """
    def __init__(self, data):
        """FreqDist Constructor

        Tokenize input data and creates a counter dictionary.

        Args:
            data (str or list of str): Text or corpus.

        Raises:
            Exception: If data is not valid.

        """
        try:
            tokens = word_tokenize(data)
        except TypeError:
            raise Exception('Input type is not supported.')

        self._counter = Counter(tokens)

    def __str__(self):
        return 'Frequency Distribution\n' + 'Most-Common: ' + \
            str(self.most_common(10))

    def __getitem__(self, key):
        return self._counter[key]

    def __setitem__(self, key, value):
        self._counter[key] = value

    def __delitem__(self, key):
        del self._counter[key]

    def get_data(self):
        """Get data method

        Converts items to tuples then returns them in a list.

        Returns:
            list of tuple: Tokens + frequencies

        """
        return [
            (key, value) for key, value in self._counter.items()
        ]

    def most_common(self, n=1):
        """Most common method

        Most common tokens and their frequencies. if n is not specified, returns
        one token.

        Args:
            n (int, optional): Number of tokens.

        Returns:
            list: Most common tokens + frequencies

        Raises:
            Exception: If n is not an integer.

        """
        if isinstance(n, int):
            return self._counter.most_common(n)
        else:
            raise Exception('N must be an integer.')

    def copy(self):
        """FreqDist Copy

        Recursively makes a deep copy of itself.

        Returns:
            FreqDist: A copy of self object

        """
        return deepcopy(self)



# -*- coding: utf-8 -*-
"""Tokenize package.

API:

* *WordTokenizer* (class): For frequent word tokenizing use with an specific
regex pattern.
* *word_tokenize* (function): For instant word tokenizing use.

* *SentTokenizer* (class): For frequent sentence tokenizing use.
* *sent_tokenize* (function): For instant use.

"""

from .word_tokenize import WordTokenizer, word_tokenize
from .sent_tokenize import SentTokenizer, sent_tokenize
